# ========================================================================
# Time::Timer - Perl code benchmarking tool
# Andrew Ho (andrew@zeuscat.com)
#
# This program contains embedded documentation in Perl POD (Plain Old
# Documentation) format. Search for the string "=head1" in this document
# to find documentation snippets, or use "perldoc" to read it; utilities
# like "pod2man" and "pod2html" can reformat as well.
#
# Copyright(c) 2000-2001 Andrew Ho.
#
# This library is free software; you can redistribute it and/or modify
# it under the same terms as Perl itself.
#
# Last modified March 23, 2001
# ========================================================================

=head1 NAME

Time::Timer - Perl code benchmarking tool

=head1 SYNOPSIS

    use Timer;
    $t = Timer->new;

    for(my $i = 0; $i < 1000; $i++) {
        $t->start('tag');
        &long_running_operation();
        $t->stop;
    }
    $t->report;

=head1 DESCRIPTION

The Timer class allows you to time portions of code conveniently, as well
as benchmark code by allowing timings of repeated trials. It is perfect
for when you need more precise information about the running time of
portions of your code than the Benchmark module will give you, but
don't want to go all out and profile your code.

The methodology is simple; create a Timer object, and wrap portions of
code that you want to benchmark with C<start()> and C<stop()> method calls.
You supply a unique tag, or event name, to those methods. This allows
one Timer object to benchmark many pieces of code.

When you have run your code (one time or over multiple trials), you can
obtain information about the running time by calling the C<results()>
method or print a descriptive benchmark report by calling C<report()>.

=head1 METHODS

=over 4

=cut


# ------------------------------------------------------------------------
# Package setup

package Time::Timer;
require 5.005;
use strict;

use Carp;
use Time::HiRes qw( gettimeofday tv_interval );

use vars qw($VERSION);
$VERSION = 0.2;


# ------------------------------------------------------------------------
# Constructor

=item $t = Timer->new;

Constructor for the Timer object; returns a reference to a Timer object.
Takes no arguments.

=cut

sub new {
    my $class = shift;
    my $self = {};
    bless $self, $class;
    return $self->reset;
}


# ------------------------------------------------------------------------
# Public methods

=item $t->reset;

Reset the Timer object to the pristine state it started in.
Erase all memory of events and any previously accumulated timings.
Returns a reference to the Timer object.

=cut

sub reset {
    my $self = shift;
    $self->{before} = {};          # [ gettimeofday ] storage
    $self->{elapsed} = {};         # elapsed fractional seconds
    $self->{lastevent} = undef;    # what the last event was
    $self->{events} = [];          # keep list of events in order seen
    return $self;
}


=item $t->start($tag);

Record the current time so that when C<stop()> is called, we can calculate
an elapsed time. Supply a $tag which is simply a string that is the
descriptive name of the event you are timing. If you do not supply a
$tag, the last event tag is used; if there is none, a "_default" tag
is used instead.

=cut

# In this routine we try hard to make the [ gettimeofday ] take place
# as late as possible to minimize Heisenberg problems. :)

sub start {
    my $self = shift;
    my $event = shift || $self->{lastevent} || '_default';
    $self->{lastevent} = $event;
    if(exists $self->{before}->{$event}) {
        push @{$self->{before}->{$event}}, [ gettimeofday ];
    } else {
        push @{$self->{events}}, $event;
        $self->{before}->{$event} = [ [ gettimeofday ] ];
    }
    return;
}


=item $t->stop($tag);

Record timing information. The optional $tag is the event for which you
are timing, and defaults to the $tag supplied to the last C<start()> call.
If a $tag is supplied, it must correspond to one given to a previously
called C<start()> call. It returns the elapsed time in milliseconds.

=cut

sub stop {
    my $after = [ gettimeofday ];    # minimize overhead
    my $self = shift;
    my $event = shift || $self->{lastevent} || '_default';

    die 'must call $t->start($event) before $t->stop($event)'
        unless exists $self->{before}->{$event};

    my $i = exists $self->{elapsed}->{$event} ?
        scalar @{$self->{elapsed}->{$event}} : 0;
    my $before = $self->{before}->{$event}->[$i];
    die 'timer out of sync' unless defined $before;

    my $elapsed = tv_interval($before, $after);
    if($i > 0) {
        push @{$self->{elapsed}->{$event}}, $elapsed;
    } else {
        $self->{elapsed}->{$event} = [ $elapsed ];
    }
    return $elapsed;
}


=item $t->report;

Print a simple report on the collected timings to STDERR. This report
prints the number of trials run, the total time taken, and, if more than
one trial was run, the average time needed to run one trial. It prints
the events out in the order they were C<start()>ed.

=cut

sub report {
    my $self = shift;
    foreach my $event (@{$self->{events}}) {
        unless(exists $self->{elapsed}->{$event}) {
            warn join ' ', 'event', $event, 'still running, skipping';
            last;
        }
        my @times = @{$self->{elapsed}->{$event}};
        my $n = scalar @times;
        my $total = 0; $total += $_ foreach @times;

        printf STDERR '%d trial%s of %s (%s total)',
            $n, $n == 1 ? '' : 's', $event, timestr($total);
        printf STDERR ', %s/trial', timestr($total / $n) if $n > 1;
        print STDERR "\n";
    }
    return;
}


=item $t->result($event);

Return the time it took for $event to elapse, or the mean time it took for
$event to elapse once, if $event happened more than once. C<result()>
will complain (via a warning) if an event is still active.

=cut

sub result {
    my $self = shift;
    my $event = shift || $self->{lastevent} || '_default';
    unless(exists $self->{elapsed}->{$event}) {
        warn join ' ', 'event', $event, 'still running';
        return;
    }
    my @times = @{$self->{elapsed}->{$event}};
    my $total = 0; $total += $_ foreach @times;
    return $total / @times;
}


=item $t->results;

Returns the timing data as a hash keyed on event tags where each value is
the time it took to run that event, or the are the average time it took,
if that event ran more than once. In scalar context it returns a reference
to that hash. The return value is actually an array, so that the original
event order is preserved.

=cut

sub results {
    my $self = shift;
    my @results;
    foreach my $event (@{$self->{events}}) {
        push @results, $event;
        push @results, $self->result($event);
    }
    return wantarray ? @results : \@results;
}


=item $t->data($event), $t->data;

These methods are useful if you want to recover the full internal timing
data to roll your own reports.

If called with an $event, returns the raw timing data for that $event as
an array (or a reference to an array if called in scalar context). This is
useful for feeding to something like the Statistics::Descriptive package.

If called with no arguments, returns the raw timing data as a hash keyed
on event tags, where the values of the hash are lists of timings for that
event. In scalar context, it returns a reference to that hash. As with
C<results()>, the data is internally represented as an array so you can
recover the original event order by assigning to an array instead of a hash.

=cut

sub data {
    my $self = shift;
    my $event = shift;
    my @results;
    if($event) {
        if(exists $self->{elapsed}->{$event}) {
            @results = @{$self->{elapsed}->{$event}};
        } else {
            @results = ();
        }
    } else {
        @results = map { ( $_ => $self->{elapsed}->{$_} || [] ) }
                          @{$self->{events}};
    }
    return wantarray ? @results : \@results;
}


# ------------------------------------------------------------------------
# Internal utility subroutines

# timestr($sec) takes a floating-point number of seconds and formats
# it in a sensible way, commifying large numbers of seconds, and
# converting to milliseconds if it makes sense. Since Time::HiRes has
# at most millisecond resolution, no attempt is made to convert into
# microseconds or below. A unit string is appended to the number.

sub timestr {
    my $sec = shift;
    my $retstr;
    if($sec >= 1_000) {
        $retstr = commify($sec) . ' s';
    } elsif($sec > 1) {
        $retstr = sprintf '%0.3f s', $sec;
    } elsif($sec == 1) {
        $retstr = '1 second';
    } elsif($sec >= 0.001) {
        $retstr = sprintf '%0.3f ms', $sec * 1_000;
    } else {
        $retstr = ($sec * 1_000) . ' ms';
    }
    $retstr;
}


# commify($num) inserts a grouping comma according to en-US standards
# for numbers larger than 1000. For example, the integer 123456 would
# be written 123,456. Any fractional part is left untouched.

sub commify {
    my $num = shift;
    return unless $num =~ /\d/;
    return $num if $num < 1_000;

    my $ip  = int $num;
    my($fp) = ($num =~ /\.(\d+)/);

    $ip =~ s/(\d\d\d)$/,$1/;
    1 while $ip =~ s/(\d)(\d\d\d),/$1,$2,/;

    return $fp ? join '.', $ip, $fp : $ip;
}


# ------------------------------------------------------------------------
# Finish up the POD.

=back

=head1 BUGS

Benchmarking is an inherently futile activity, fraught with uncertainty
not dissimilar to that experienced in quantum mechanics.

=head1 SEE ALSO

L<Benchmark>, L<Time::HiRes>, L<Time::Stopwatch>, L<Statistics::Descriptive>

=head1 AUTHOR

Andrew Ho E<lt>andrew@zeuscat.comE<gt>

=head1 COPYRIGHT

Copyright(c) 2000-2001 Andrew Ho.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut


# ------------------------------------------------------------------------
# Return true for a valid Perl include

1;


# ========================================================================
__END__
